/* CGV/solution/parts/cgvp00.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Stephan Scheiderer                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 00 <<<<<<<<<<<<<<<<<<<<<<*/

#include <stdio.h>
#include <stdlib.h>
#ifndef serial
#  include <mpi.h>
#else
#  include <time.h>
#endif
#include <math.h>
#include <malloc.h>
#include <string.h>

/****************** declarations and global data ***************/

#ifndef serial
#  define TRUE    1
#  define FALSE   0
#  define MSG_TAG 333
#endif

/* options: */
int     m,n;         /* size of the physical-heat matrix without borders */
                     /* m=vertical, n=horizontal */
int     iter_max;    /* maximal number of iterations in CG */
double  epsilon;     /* Limit for sqr( || residual vector || ) */
int     print_level; /* print_level = 1: Only result-error and partial result matrix */
                                 /* = 2: and residual norm after each iteration */
                                 /* = 3: and result physical-heat matrix */
                                 /* = 4: and all vector and halo data in first iteration */
                                 /* = 5: and all vector and halo data in _all_ iterations */
int     decomp_dims; /* dimensins of the domain decomposition */
int     m_procs, n_procs; /* number of processes in both directions of the 2-dim domain decomposition */
int     mprt, nprt;  /* number of rows and columns when printing the partial result matrix */

/* global structures and data: */

#ifndef serial
 MPI_Comm comm_cart;
#endif
int numprocs, my_rank;
unsigned long flops;

/* size of the physical area this process is about to calculate
 * start_i   : start row
 * end1_i    : end row + 1
 * start_j   : start column
 * end1_j    : end column + 1
 * rowsize   : number of data entries in a row of the local chunk
 *             (end1_j - start_j)
 * colsize   : number of data entries in a column of the local chunk
 *             (end1_i - start_i)
 * chunksize : number of data entries in the local chunk
 *             (rowsize * colsize)
 */
int start_i, end1_i, start_j, end1_j, rowsize, colsize, chunksize;

/* SIDX (Serial InDeX): computes the serial index of a two-dimensional array */
/*   input: i = 0...(colsize - 1) and j = 0...(rowsize - 1) */
/*   output: SIDX in 0..(colsize*rowsize-1) */
#  define SIDX(i, j) (rowsize * (i) + (j))

/* data structures */

/* chunk of a vector */
typedef double* vector_chunk;
vector_chunk x, b;


#ifndef serial
/* halo structure */
struct halo_struct
{
    double* north;
    int     north_rank;
    int     north_size;
    double* east;
    int     east_rank;
    int     east_size;
    double* south;
    int     south_rank;
    int     south_size;
    double* west;
    int     west_rank;
    int     west_size;
};
struct halo_struct halo;

/* send scratch buffer */
double* send_scratch_buff;
#endif

/********************** memory management **********************/

void abrt(void)
{
#ifndef serial
    MPI_Abort(MPI_COMM_WORLD, -1);
#endif
    exit(-1);
}

/* safe_malloc: Same meaning as malloc(); error handling included */
void* safe_malloc(int size)
{
    void *new;

    if (size == 0) return NULL;

    new = (void*)malloc(size);
    if (new == NULL) {
        printf("Could not allocate memory.\n");
        abrt();
    } 
    return(new);
}

/********************** commandline options ********************/

void read_options( int argc, char* argv[] )
{
    int optid = 1; /* start at first optional argument */
    int err_detected = 0;
    if ( my_rank == 0 ) {
        while ( optid < argc ) {
            if (!strcmp( argv[optid], "-m") ) { optid++;
                m        = atoi( argv[optid] ); if (m<=0) {err_detected = 1; break;}
            } else if (!strcmp( argv[optid], "-n") ) { optid++;
                n        = atoi( argv[optid] ); if (n<=0) {err_detected = 1; break;}
            } else if (!strcmp( argv[optid], "-imax") ) { optid++;
                iter_max = atoi( argv[optid] ); if (iter_max<=0) {err_detected = 1; break;}
            } else if (!strcmp( argv[optid], "-eps") ) { optid++;
                epsilon  = atof( argv[optid] ); if (epsilon<=0) {err_detected = 1; break;}
            } else if (!strcmp( argv[optid], "-prtlev") ) { optid++;
                print_level = atoi( argv[optid] ); if (print_level<0) {err_detected = 1; break;}
            } else if (!strcmp( argv[optid], "-twodims") ) {
                decomp_dims = 2;
            } else if (!strcmp( argv[optid], "-mprocs") ) { optid++;
                m_procs = atoi( argv[optid] ); if (m_procs<=0) {err_detected = 1; break;}
            } else if (!strcmp( argv[optid], "-nprocs") ) { optid++;
                n_procs = atoi( argv[optid] ); if (n_procs<=0) {err_detected = 1; break;}
            } else {err_detected = 1; break;}
            optid++;
        }
        if ( err_detected ) {
            printf( "unknown command line option: %s\n", argv[optid] );
            printf( "\n");
            printf( "Synopsis: mpirun -np <number of processes> ./cg_all <options>\n");
            printf( "  with options [defaults]:\n");
            printf( "  -m <number of rows in physical area of the Laplace equation> [4]\n");
            printf( "  -n <number of columns in physical area of the Laplace equation [4]\n");
            printf( "  -imax <maximal number of iterations> [500]\n");
            printf( "  -eps  <epsion> [1e-6]\n");
            printf( "  -prtlev <print and debug level 0..5> [1]\n");
            printf( "  -twodims [only one-dimensional domain decomposition (DD)]\n");
            printf( "  -mprocs <horizontal number of processes in 2-dim DD\n");
            printf( "  -nprocs <vertical   number of processes in 2-dim DD\n");
            printf( "\n");
            fflush(stdout);
            abrt();
        }
    }

#ifndef serial
    MPI_Bcast(&m,          1, MPI_INT,    0, MPI_COMM_WORLD);
    MPI_Bcast(&n,          1, MPI_INT,    0, MPI_COMM_WORLD);
    MPI_Bcast(&iter_max,   1, MPI_INT,    0, MPI_COMM_WORLD);
    MPI_Bcast(&epsilon,    1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&print_level,1, MPI_INT,    0, MPI_COMM_WORLD);
    MPI_Bcast(&decomp_dims,1, MPI_INT,    0, MPI_COMM_WORLD);
    MPI_Bcast(&m_procs,    1, MPI_INT,    0, MPI_COMM_WORLD);
    MPI_Bcast(&n_procs,    1, MPI_INT,    0, MPI_COMM_WORLD);
#endif
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 00 <<<<<<<<<<<<<<<<<<<<<<*/
